/* Name: main.c
 * Project: hid-data, example how to use HID for data transfer
 * Author: Christian Starkjohann   edited by Fred Froehlich
 * Creation Date: 2008-04-11
 * Copyright: (c) 2008 by OBJECTIVE DEVELOPMENT Software GmbH
 * License: GNU GPL v2 (see License.txt) or proprietary (CommercialLicense.txt)
 */


#include <stdio.h>
#include <string.h>
#include <avr/pgmspace.h>

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>  /* for sei() */
#include <util/delay.h>     /* for _delay_ms() */

#include <avr/pgmspace.h>   /* required by usbdrv.h */
#include "usbdrv/usbdrv.h"
#include "i2c/hw_i2c.h"
#include <avr/eeprom.h> 
#include "main.h"



/* ------------------------------------------------------------------------- */
/* ----------------------------- USB interface ----------------------------- */
/* ------------------------------------------------------------------------- */

//Verwendete Befehle im Flash ablegen
char p_info[] 	PROGMEM = "INFO";
char p_read[] 	PROGMEM = "READ";
char p_chck[] 	PROGMEM = "CHCK";
char p_writ[] 	PROGMEM = "WRIT";
char p_wait[] 	PROGMEM = "WAIT";
char p_firm[] 	PROGMEM = "FIRM";

// Befehls-Tabelle (im Flash) und Funktionspointer
CMD_ITEM CMD_TABLE[] PROGMEM = 
{
	{p_info,cmd_info},
	{p_read,cmd_read},
	{p_chck,cmd_chck},
	{p_writ,cmd_writ},
	{p_wait,cmd_wait},
	{p_firm,cmd_firm},
	{NULL,NULL} 
};

// USB report descriptor 
PROGMEM char usbHidReportDescriptor[22] = 
{    
    0x06, 0x00, 0xff,              // USAGE_PAGE (Generic Desktop)
    0x09, 0x01,                    // USAGE (Vendor Usage 1)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x26, 0xff, 0x00,              //   LOGICAL_MAXIMUM (255)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x95, 0x80,                    //   REPORT_COUNT (128)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x02, 0x01,              //   FEATURE (Data,Var,Abs,Buf)
    0xc0                           // END_COLLECTION
};

/* Since we define only one feature report, we don't use report-IDs (which
 * would be the first byte of the report). The entire report consists of 128
 * opaque data bytes.
 */

/* The following variables store the status of the current data transfer */
static unsigned char    currentAddress;
static unsigned char    bytesRemaining;

// -------------------------------------------------------------------------
// Adapter ---->> PC
uchar   usbFunctionRead(uchar *data, uchar len)
{
    if(len > bytesRemaining) len = bytesRemaining;
    for(unsigned char i = 0; i < len; i++)
        data[i] = usb_send_buffer[currentAddress+i];

    currentAddress += len;
    bytesRemaining -= len;
	
    return len;
}

// -------------------------------------------------------------------------
// PC ---->> Adapter
uchar   usbFunctionWrite(uchar *data, uchar len)
{
    if(bytesRemaining == 0) return 1;               // end of transfer
    if(len > bytesRemaining) len = bytesRemaining;
   
    for(unsigned char i=0;i<len; i++)
	{
		usb_rcv_buffer[currentAddress+i] = data[i];
	}
	currentAddress += len;
    bytesRemaining -= len;
	if(bytesRemaining == 0)
	{
		for(uint8_t i = 0; i<USB_BUFFER_SIZE; i++)
			usb_send_buffer[i] = 0;		//Daten noch nicht fr neusten Befehl aufbereitet
		g_readgoon = 0;
		getcommand();	//
	}
    return bytesRemaining == 0; // return 1 if this was the last chunk 
}

// -------------------------------------------------------------------------
usbMsgLen_t usbFunctionSetup(uchar data[8])
{
usbRequest_t    *rq = (void *)data;

    if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){    // HID class request
		LED_GR_AUS;
        if(rq->bRequest == USBRQ_HID_GET_REPORT){  // wValue: ReportType (highbyte), ReportID (lowbyte)
            // since we have only one report type, we can ignore the report-ID
				bytesRemaining = USB_BUFFER_SIZE -1;
				currentAddress = 0;
				g_readready4next = 1;
			return USB_NO_MSG;  // use usbFunctionRead() to obtain data
			
        }
		else if(rq->bRequest == USBRQ_HID_SET_REPORT)
		{
            // since we have only one report type, we can ignore the report-ID
            bytesRemaining = USB_BUFFER_SIZE -1;
            currentAddress = 0;
			return USB_NO_MSG;  // use usbFunctionWrite() to receive data from host
        }
    }else{
        // ignore vendor type requests, we don't use any
    }
    return 0;
}

/* ------------------------------------------------------------------------- */

int main(void)
{
	unsigned char   i;

    wdt_enable(WDTO_2S);
    /* Even if you don't use the watchdog, turn it off here. On newer devices,
     * the status of the watchdog (on/off, period) is PRESERVED OVER RESET!
     */

    /* RESET status: all port bits are inputs without pull-up.
     * That's the way we need D+ and D-. Therefore we don't need any
     * additional hardware initialization.
     */

	//Konfiguration der PORTS 
	//1=OUTPUT / 0=INPUT
//	DDRA = 0b00000000;
	DDRB = 0b00000000;
	DDRC = 0b00001100;
	DDRD = 0b00000000;

	PORTB = 1 << 4; 			//PullUp fr ResetTaster

	LED_GR_AUS;
	LED_GE_AUS;

	i2c_init();
	
	for(i=0;i<129;i++) 
	{
		usb_rcv_buffer[i] = 0;
		usb_send_buffer[i] = 0;
	}
	
    usbInit();
    usbDeviceDisconnect();  /* enforce re-enumeration, do this while interrupts are disabled! */
    i = 0;
    while(--i){             /* fake USB disconnect for > 250 ms */
        wdt_reset();
        _delay_ms(2);
    }
    usbDeviceConnect();
	LED_GR_AUS;
	
	usbInit(); 
    sei();

    for(;;){                /* main event loop */
        wdt_reset();
        usbPoll();
		
		if((bytesRemaining == 0) && (g_readgoon == 1) && (g_readready4next==1))
		{
			cmd_read();
		}
		
		LED_GR_EIN;
		want_reset();		
    }
    return 0;
}

// ------------------------------------------------------------------------------------------------------------------- 
void want_reset()
{ 
//	wdt_enable(WDTO_2S);
    for(;;)
    {
		if((PINB & (1 << 4)) != 0)   return;
		LED_GR_EIN;
		LED_GE_EIN;
    }           
}

// -------------------------------------------------------------------------------------------------------------------
void getcommand()
{
	char cmd_t[5];
	unsigned char cmd_index = 0;
 
	//Commando (erste 4 Bytes des USB Buffers separieren
	cmd_t[0] = usb_rcv_buffer[0];
	cmd_t[1] = usb_rcv_buffer[1];
	cmd_t[2] = usb_rcv_buffer[2];
	cmd_t[3] = usb_rcv_buffer[3];
	cmd_t[4] = 0;	//Stringende

	//Kommando in Tabelle CMD_TABLE suchen
	while(strcasecmp_P(&cmd_t[0],(char *)pgm_read_word(&CMD_TABLE[cmd_index].cmd)))
    {
        // wenn letzter Eintrag ({NULL,NULL}) erreicht wurde,
		// Abruch der Whileschleife und Unterprogramm verlassen 
        if (pgm_read_word(&CMD_TABLE[++cmd_index].cmd) == 0) 
		{
			error();
			return;
		}
    }
  	// der Befehl wurde in der Tabelle gefunden;
	((cmd_fp)pgm_read_word(&CMD_TABLE[cmd_index].fp))(); // Befehl ausfhren
	return; 
}

// -------------------------------------------------------------------------------------------------------------------
void error()	//Sendet eine Errormeldung zurck
{
	usb_send_buffer[0] = 'E';
	usb_send_buffer[1] = 'R';
	usb_send_buffer[2] = 'R';
	usb_send_buffer[3] = ':';
	usb_send_buffer[4] = usb_rcv_buffer[0];
	usb_send_buffer[5] = usb_rcv_buffer[1];
	usb_send_buffer[6] = usb_rcv_buffer[2];
	usb_send_buffer[7] = usb_rcv_buffer[3];
	usb_send_buffer[8] = 0;
}

// -------------------------------------------------------------------------------------------------------------------
void ok()	//Sendet eine Errormeldung zurck
{
	usb_send_buffer[0] = '!';
	usb_send_buffer[1] = 'O';
	usb_send_buffer[2] = 'K';
	usb_send_buffer[3] = ':';
	usb_send_buffer[4] = usb_rcv_buffer[0];
	usb_send_buffer[5] = usb_rcv_buffer[1];
	usb_send_buffer[6] = usb_rcv_buffer[2];
	usb_send_buffer[7] = usb_rcv_buffer[3];
	usb_send_buffer[8] = 0;
}

// -------------------------------------------------------------------------------------------------------------------
//Sendet Daten ber den Adapter (FWVersion,HWVersion,ProzessorTakt,TWiPrescalar,TWIBitrate,Compilerversion + Datum-Zeit)
void cmd_info()	
{
	//AN DIESER STRUKTUR NICHTS NDERN -> CAMFRED NUTZT DAS ZUR AUSWERTUNG
	char i = TWSR;
	char i1 = TWBR;
	i &= 3;
	
	sprintf_P((char *)&usb_send_buffer[16],PSTR("FW %i/HW %i/PC %lu/TC %u/TP %i/GC %s/DT %s/TM %s"),fw_version, hw_version,F_CPU,i1,i,__VERSION__,__DATE__, __TIME__);

	ok();
}

// -------------------------------------------------------------------------------------------------------------------
void cmd_read()
{
    unsigned char i, count=0;
	unsigned char adrH;
	unsigned char adrL;
	
	LED_GE_EIN;
	g_readready4next=0;

	if(!g_readgoon)	//erster Datensatz Startadresse und Anzahl sichern
	{
		g_chipid = usb_rcv_buffer[8];		//0 bei 64K CamPacs ansonsten 160
		g_readcount = (usb_rcv_buffer[6]<<8) + usb_rcv_buffer[7];
		g_readadres = (usb_rcv_buffer[4]<<8) + usb_rcv_buffer[5];
	}
	else
	{
		for(uint8_t i = 0; i<USB_BUFFER_SIZE; i++)
			usb_send_buffer[i] = 0;		//Daten lschen
	}
	g_readgoon = 1;
		
	usb_send_buffer[10] = g_readadres >> 8;		//erste Adresse der CamPac Bytes im Datensatz
	usb_send_buffer[11] = g_readadres & 255;
	
	while(count<64)
	{
		adrH = g_readadres >> 8;
		adrH = adrH << 1;
		adrL = g_readadres & 255;
	
		i = i2c_start_wait(g_chipid + adrH + I2C_WRITEFLAG,250);       //160 + Block zur Addressierung
		wdt_reset();
		if(g_chipid < 160) _delay_us(250);	//kurze Pause bei 64K CamPacs damit der C die 2. Adresse gereieren kann 
		if (i)		//Fehler 
		{
			i2c_stop();
			error();
			LED_GE_AUS;
			g_readgoon = 0;
			return;
		}

		i2c_write(adrL);
		i = i2c_start_wait(g_chipid + adrH + I2C_READFLAG,250);       // set device address and read mode
		if(g_chipid < 160) _delay_us(250);	//kurze Pause bei 64K CamPacs damit der C die 2. Adresse gereieren kann 
		wdt_reset();
		if (i)		//Fehler 
		{
			i2c_stop();
			error();
			LED_GE_AUS;
			g_readgoon = 0;
			return;
		}

		usb_send_buffer[16+count] = i2c_readNak();                    // read one byte
		i2c_2ndack();
//		i2c_stop();
		
		count++;
		g_readadres++;
		g_readcount--;
		if(g_readcount == 0)
		{
			g_readgoon = 0;
			break;
		}
	}
	usb_send_buffer[8] = 0;		//Anzahl der CamPac Bytes im Datensatz
	usb_send_buffer[9] = count;
	
	ok();
	

	LED_GE_AUS;
}

// -------------------------------------------------------------------------------------------------------------------
void cmd_chck()
{
    unsigned char i,i1,size = 16;
	LED_GE_EIN;

	while(size>0)
	{
		if(size==16) i = i2c_start(62);	//64K CamPac
		else i = i2c_start(160 | ((size-1)*2));
		_delay_ms(1);
		i1 = i2c_write(255);	
		i2c_stop();

		if(i || i1)	//Keine Antwort -> kleineres Modul gesteckt
		{
			size = size >> 1;
			wdt_reset();
		}
		else	//Antwort erhalten -> Modul dieser Grsse ist ansprechbar
		{
			if(size == 16) size = size * 2;	//64K Modul
			usb_send_buffer[16+i] = size;
			break;
		}
	}
	ok();
	LED_GE_AUS;
}

// -------------------------------------------------------------------------------------------------------------------
void cmd_writ()
{
    unsigned char i, count=0, anzahlbytes;
	unsigned int adresse;
	unsigned char adrH;
	unsigned char adrL;
	
	LED_GE_EIN;

	g_chipid = usb_rcv_buffer[8];		//0 bei 64K CamPacs ansonsten 160
	anzahlbytes = (usb_rcv_buffer[6]<<8) + usb_rcv_buffer[7];		//Anzahl der zu schreibenden Bytes
	adresse = (usb_rcv_buffer[4]<<8) + usb_rcv_buffer[5];			//Startadresse
			
	while(count<64)
	{
		adrH = adresse >> 8;
		adrH = adrH << 1;
		adrL = adresse & 255;
	
		i = i2c_start_wait(g_chipid + adrH + I2C_WRITEFLAG,250);       //160 + Block zur Addressierung
		if(g_chipid < 160) _delay_us(250);	//kurze Pause bei 64K CamPacs damit der C die 2. Adresse gereieren kann 
		wdt_reset();
		if (i)		//Fehler 
		{
			i2c_stop();
			error();
			LED_GE_AUS;
			return;
		}

		i = i2c_write(adrL);						//Setze Adresse
		if (i)		//Fehler 
		{
			i2c_stop();
			error();
			LED_GE_AUS;
			return;
		}
		i = i2c_write(usb_rcv_buffer[16+count]);	//schreibe Wert
		if (i)		//Fehler 
		{
			i2c_stop();
			error();
			LED_GE_AUS;
			return;
		}
		if(g_chipid < 160) _delay_ms(3);	//kurze Pause bei 64K CamPacs falls Clockverlngerung nicht programmiert
		i2c_stop();
		
		count++;
		adresse++;
		anzahlbytes--;
		if(anzahlbytes == 0)
		{
			break;
		}
	}
	usb_send_buffer[8] = 0;		//Anzahl der geschriebenen Bytes 
	usb_send_buffer[9] = count;
	
	ok();
	LED_GE_AUS;
}

// -------------------------------------------------------------------------------------------------------------------
void cmd_wait()		//Setzt die Taktfrequenz fr den TWI Bus
{
	//SCLFreq = F_CPU/ 16 + 2(TWBR) * (4^TWSR)

	if(usb_rcv_buffer[4] > 3)		//Prescalar: 0=1; 1=4; 2=16; 3=64 Teiler 
	{
		error();
		return;
	}
	if(usb_rcv_buffer[5] < 10)		//must be > 10 for stable operation
	{
		error();
		return;
	}
	
	TWSR = usb_rcv_buffer[4];		//Prescalar (1,4,16,64 mglich)
	TWBR = usb_rcv_buffer[5];  		//must be > 10 for stable operation

	//TWI[6] muss 5 sein, damit Daten als gltig angesehen werden, sonst erfolgt Einstellung nur temporr (siehe I2C_Init)
	eeprom_write_block(&usb_rcv_buffer[4], (uchar *)0, 3); 

	ok();
}

// -------------------------------------------------------------------------------------------------------------------
void cmd_firm(){}
